"""High-level reliable submission methods with XRPL transactions."""

import asyncio
from typing import Optional, Union

from typing_extensions import Final

from xrpl.asyncio.clients import Client
from xrpl.asyncio.ledger import get_latest_validated_ledger_sequence
from xrpl.asyncio.transaction.main import _check_fee
from xrpl.asyncio.transaction.main import autofill as _autofill
from xrpl.asyncio.transaction.main import sign, submit
from xrpl.clients import XRPLRequestFailureException
from xrpl.constants import XRPLException
from xrpl.models.requests import Tx
from xrpl.models.response import Response
from xrpl.models.transactions.transaction import Transaction
from xrpl.wallet.main import Wallet

_LEDGER_CLOSE_TIME: Final[int] = 1


class XRPLReliableSubmissionException(XRPLException):
    """General XRPL Reliable Submission Exception."""

    pass


async def _wait_for_final_transaction_outcome(
    transaction_hash: str, client: Client, prelim_result: str, last_ledger_sequence: int
) -> Response:
    """
    The core logic of reliable submission.  Polls the ledger until the result of the
    transaction can be considered final, meaning it has either been included in a
    validated ledger, or the transaction's LastLedgerSequence has been surpassed by the
    latest ledger sequence (meaning it will never be included in a validated ledger).
    """
    await asyncio.sleep(_LEDGER_CLOSE_TIME)

    current_ledger_sequence = await get_latest_validated_ledger_sequence(client)

    if current_ledger_sequence >= last_ledger_sequence:
        raise XRPLReliableSubmissionException(
            f"The latest validated ledger sequence {current_ledger_sequence} is "
            f"greater than LastLedgerSequence {last_ledger_sequence} in "
            f"the transaction. Prelim result: {prelim_result}"
        )

    # query transaction by hash
    transaction_response = await client._request_impl(Tx(transaction=transaction_hash))
    if not transaction_response.is_successful():
        if transaction_response.result["error"] == "txnNotFound":
            """
            For the case if a submitted transaction is still
            in queue and not processed on the ledger yet.
            """
            return await _wait_for_final_transaction_outcome(
                transaction_hash, client, prelim_result, last_ledger_sequence
            )
        else:
            raise XRPLRequestFailureException(transaction_response.result)

    result = transaction_response.result
    if "validated" in result and result["validated"]:
        # result is in a validated ledger, outcome is final
        return_code = result["meta"]["TransactionResult"]
        if return_code != "tesSUCCESS":
            raise XRPLReliableSubmissionException(f"Transaction failed: {return_code}")
        return transaction_response

    # outcome is not yet final
    return await _wait_for_final_transaction_outcome(
        transaction_hash, client, prelim_result, last_ledger_sequence
    )


async def _send_reliable_submission(
    transaction: Transaction, client: Client, *, fail_hard: bool = False
) -> Response:
    """
    Asynchronously submits a transaction and verifies that it has been included in a
    validated ledger (or has errored/will not be included for some reason).

    `See Reliable Transaction Submission
    <https://xrpl.org/reliable-transaction-submission.html>`_

    Note: This cannot be used with a standalone rippled node, because ledgers do not
    close automatically.

    Args:
        transaction: the signed transaction to submit to the ledger. Requires a
            `last_ledger_sequence` param.
        client: the network client used to submit the transaction to a rippled node.
        fail_hard: an optional boolean. If True, and the transaction fails for
            the initial server, do not retry or relay the transaction to other
            servers. Defaults to False.

    Returns:
        The response from a validated ledger.

    Raises:
        XRPLReliableSubmissionException: if the transaction fails, is malformed, or is
            missing a `last_ledger_sequence` param.
    """
    if transaction.last_ledger_sequence is None:
        raise XRPLReliableSubmissionException(
            "Transaction must have a `last_ledger_sequence` param."
        )
    transaction_hash = transaction.get_hash()
    submit_response = await submit(transaction, client, fail_hard=fail_hard)
    prelim_result = submit_response.result["engine_result"]
    if prelim_result[0:3] == "tem":
        error_message = submit_response.result["engine_result_message"]
        raise XRPLReliableSubmissionException(f"{prelim_result}: {error_message}")

    return await _wait_for_final_transaction_outcome(
        transaction_hash, client, prelim_result, transaction.last_ledger_sequence
    )


async def _get_signed_tx(
    transaction: Union[Transaction, str],
    client: Client,
    wallet: Optional[Wallet] = None,
    check_fee: bool = True,
    autofill: bool = True,
) -> Transaction:
    """
    Sets up a transaction to be submitted by optionally autofilling, signing,
        and checking the fee validity.

    Args:
        transaction: the transaction or transaction blob to be signed if unsigned.
        client: the network client to autofill from.
        wallet: an optional wallet with which to sign the transaction. This is
            only needed if the transaction is unsigned.
        check_fee: an optional bolean indicating whether to check if the fee is
            higher than the expected transaction type fee. Defaults to True.
        autofill: an optional boolean indicating whether to autofill the
            transaction. Defaults to True.

    Returns:
        The signed transaction.

    Raises:
        XRPLException: if the transaction is unsigned and a wallet is not
            provided for signing.
    """
    if isinstance(transaction, str):
        transaction = Transaction.from_blob(transaction)

    if transaction.is_signed():
        return transaction

    # Attempts to validate fee, autofill, and sign the transaction when unsigned.
    if not wallet:
        raise XRPLException(
            "Wallet must be provided when submitting an unsigned transaction"
        )

    if check_fee:
        await _check_fee(transaction, client)

    if autofill:
        transaction = await _autofill(transaction, client)

    if transaction.signers:
        return sign(transaction, wallet, multisign=True)

    return sign(transaction, wallet, multisign=False)


async def submit_and_wait(
    transaction: Union[Transaction, str],
    client: Client,
    wallet: Optional[Wallet] = None,
    *,
    check_fee: bool = True,
    autofill: bool = True,
    fail_hard: bool = False,
) -> Response:
    """
    Signs a transaction locally if the transaction is unsigned, then submits,
    and verifies that it has been included in a validated ledger (or has errored/
    will not be included for some reason).
    `See Reliable Transaction Submission for a full explanation of this workflow
    <https://xrpl.org/reliable-transaction-submission.html>`_

    Args:
        transaction: the signed/unsigned transaction (or transaction blob) to
            be submitted.
        client: the network client with which to submit the transaction.
        wallet: an optional wallet with which to sign the transaction. This is
            only needed if the transaction is unsigned.
        check_fee: an optional bolean indicating whether to check if the fee is
            higher than the expected transaction type fee. Defaults to True.
        autofill: an optional boolean indicating whether to autofill the
            transaction. Defaults to True.
        fail_hard: an optional boolean. If True, and the transaction fails for
            the initial server, do not retry or relay the transaction to other
            servers. Defaults to False.

    Returns:
        The response from the ledger.
    """
    signed_transaction = await _get_signed_tx(
        transaction, client, wallet, check_fee, autofill
    )
    return await _send_reliable_submission(
        signed_transaction, client, fail_hard=fail_hard
    )



"""
The subscribe method requests periodic notifications from the server
when certain events happen.

WebSocket API only.

`See subscribe <https://xrpl.org/subscribe.html>`_
"""
from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional

from xrpl.models.base_model import BaseModel
from xrpl.models.currencies import Currency
from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import require_kwargs_on_init


class StreamParameter(str, Enum):
    """Represents possible values of the streams query param for subscribe."""

    CONSENSUS = "consensus"
    LEDGER = "ledger"
    MANIFESTS = "manifests"
    PEER_STATUS = "peer_status"
    TRANSACTIONS = "transactions"
    TRANSACTIONS_PROPOSED = "transactions_proposed"
    SERVER = "server"
    VALIDATIONS = "validations"


@require_kwargs_on_init
@dataclass(frozen=True)
class SubscribeBook(BaseModel):
    """Format for elements in the ``books`` array for Subscribe only."""

    taker_gets: Currency = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    taker_pays: Currency = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    taker: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    snapshot: bool = False
    both: bool = False


@require_kwargs_on_init
@dataclass(frozen=True)
class Subscribe(Request):
    """
    The subscribe method requests periodic notifications from the server
    when certain events happen.

    WebSocket API only.

    `See subscribe <https://xrpl.org/subscribe.html>`_
    """

    method: RequestMethod = field(default=RequestMethod.SUBSCRIBE, init=False)
    streams: Optional[List[StreamParameter]] = None
    accounts: Optional[List[str]] = None
    accounts_proposed: Optional[List[str]] = None
    books: Optional[List[SubscribeBook]] = None
    url: Optional[str] = None
    url_username: Optional[str] = None
    url_password: Optional[str] = None



"""Model for AMMDeposit transaction type."""
from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Optional

from xrpl.models.amounts import Amount, IssuedCurrencyAmount
from xrpl.models.currencies import Currency
from xrpl.models.flags import FlagInterface
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init


class AMMDepositFlag(int, Enum):
    """
    Transactions of the AMMDeposit type support additional values in the Flags field.
    This enum represents those options.
    """

    TF_LP_TOKEN = 0x00010000
    TF_SINGLE_ASSET = 0x00080000
    TF_TWO_ASSET = 0x00100000
    TF_ONE_ASSET_LP_TOKEN = 0x00200000
    TF_LIMIT_LP_TOKEN = 0x00400000


class AMMDepositFlagInterface(FlagInterface):
    """
    Transactions of the AMMDeposit type support additional values in the Flags field.
    This TypedDict represents those options.
    """

    TF_LP_TOKEN: bool
    TF_SINGLE_ASSET: bool
    TF_TWO_ASSET: bool
    TF_ONE_ASSET_LP_TOKEN: bool
    TF_LIMIT_LP_TOKEN: bool


@require_kwargs_on_init
@dataclass(frozen=True)
class AMMDeposit(Transaction):
    """
    Deposit funds into an Automated Market Maker (AMM) instance
    and receive the AMM's liquidity provider tokens (LP Tokens) in exchange.

    You can deposit one or both of the assets in the AMM's pool.
    If successful, this transaction creates a trust line to the AMM Account (limit 0)
    to hold the LP Tokens.
    """

    asset: Currency = REQUIRED  # type: ignore
    """
    The definition for one of the assets in the AMM's pool. This field is required.
    """

    asset2: Currency = REQUIRED  # type: ignore
    """
    The definition for the other asset in the AMM's pool. This field is required.
    """

    amount: Optional[Amount] = None
    """
    The amount of one asset to deposit to the AMM.
    If present, this must match the type of one of the assets (tokens or XRP)
    in the AMM's pool.
    """

    amount2: Optional[Amount] = None
    """
    The amount of another asset to add to the AMM.
    If present, this must match the type of the other asset in the AMM's pool
    and cannot be the same asset as Amount.
    """

    e_price: Optional[Amount] = None
    """
    The maximum effective price, in the deposit asset, to pay
    for each LP Token received.
    """

    lp_token_out: Optional[IssuedCurrencyAmount] = None
    """
    How many of the AMM's LP Tokens to buy.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.AMM_DEPOSIT,
        init=False,
    )

    def _get_errors(self: AMMDeposit) -> Dict[str, str]:
        errors = super()._get_errors()
        if self.amount2 is not None and self.amount is None:
            errors["AMMDeposit"] = "Must set `amount` with `amount2`"
        elif self.e_price is not None and self.amount is None:
            errors["AMMDeposit"] = "Must set `amount` with `e_price`"
        elif self.lp_token_out is None and self.amount is None:
            errors["AMMDeposit"] = "Must set at least `lp_token_out` or `amount`"
        return errors



"""Model for AMMCreate transaction type."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, Optional

from typing_extensions import Final

from xrpl.models.amounts import Amount
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init

AMM_MAX_TRADING_FEE: Final[int] = 1000


@require_kwargs_on_init
@dataclass(frozen=True)
class AMMCreate(Transaction):
    """
    Create a new Automated Market Maker (AMM) instance for trading a pair of
    assets (fungible tokens or XRP).

    Creates both an AMM object and a special AccountRoot object to represent the AMM.
    Also transfers ownership of the starting balance of both assets from the sender to
    the created AccountRoot and issues an initial balance of liquidity provider
    tokens (LP Tokens) from the AMM account to the sender.

    Caution: When you create the AMM, you should fund it with (approximately)
    equal-value amounts of each asset.
    Otherwise, other users can profit at your expense by trading with
    this AMM (performing arbitrage).
    The currency risk that liquidity providers take on increases with the
    volatility (potential for imbalance) of the asset pair.
    The higher the trading fee, the more it offsets this risk,
    so it's best to set the trading fee based on the volatility of the asset pair.
    """

    amount: Amount = REQUIRED  # type: ignore
    """
    The first of the two assets to fund this AMM with. This must be a positive amount.
    This field is required.
    """

    amount2: Amount = REQUIRED  # type: ignore
    """
    The second of the two assets to fund this AMM with. This must be a positive amount.
    This field is required.
    """

    trading_fee: int = REQUIRED  # type: ignore
    """
    The fee to charge for trades against this AMM instance, in units of 1/100,000;
    a value of 1 is equivalent to 0.001%.
    The maximum value is 1000, indicating a 1% fee.
    The minimum value is 0.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.AMM_CREATE,
        init=False,
    )

    def _get_errors(self: AMMCreate) -> Dict[str, str]:
        return {
            key: value
            for key, value in {
                **super()._get_errors(),
                "trading_fee": self._get_trading_fee_error(),
            }.items()
            if value is not None
        }

    def _get_trading_fee_error(self: AMMCreate) -> Optional[str]:
        if self.trading_fee < 0 or self.trading_fee > AMM_MAX_TRADING_FEE:
            return f"Must be between 0 and {AMM_MAX_TRADING_FEE}"
        return None
